//
//  StepperController.m
//  Stepper
//
//  Created by Phidgets on 2015-07-22.
//  Copyright (c) 2015 Phidgets. All rights reserved.
//

#import "StepperController.h"

@implementation StepperController

- (void)windowWillClose:(NSNotification*)aNotification{
    if(ch){
        //Safely close example
        Phidget_setOnAttachHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_setOnDetachHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_setOnErrorHandler((PhidgetHandle)ch, NULL, NULL);
        Phidget_close((PhidgetHandle)ch);
        PhidgetStepper_delete(&ch);
        ch = NULL;
    }
}

-(void)awakeFromNib{
    PhidgetReturnCode result;
    
    //Manage GUI
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
    [mainWindow setDelegate:self];
    [phidgetInfoBoxView fillPhidgetInfo:nil];
    [phidgetInfoBox setHidden:NO];
    
    result = PhidgetStepper_create(&ch);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    
    result = [self initChannel:(PhidgetHandle)ch];
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    
    result = PhidgetStepper_setOnStoppedHandler(ch, gotStoppedData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    
    result = PhidgetStepper_setOnPositionChangeHandler(ch, gotPositionChangeData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    
    result = PhidgetStepper_setOnVelocityChangeHandler(ch, gotVelocityChangeData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    
    [PhidgetInfoBox openCmdLine:(PhidgetHandle)ch];
    [tabBox setDelegate:self];
}

-(PhidgetReturnCode)initChannel:(PhidgetHandle) channel{
    PhidgetReturnCode result;
    
    result = Phidget_setOnAttachHandler(channel, gotAttach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    result = Phidget_setOnDetachHandler(channel, gotDetach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    result = Phidget_setOnErrorHandler(channel, gotError, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        return result;
    }
    
    /*
     * Please review the Phidget22 channel matching documentation for details on the device
     * and class architecture of Phidget22, and how channels are matched to device features.
     */
    
    /*
     * Specifies the serial number of the device to attach to.
     * For VINT devices, this is the hub serial number.
     *
     * The default is any device.
     */
    // Phidget_setDeviceSerialNumber(ch, <YOUR DEVICE SERIAL NUMBER>);
    
    /*
     * For VINT devices, this specifies the port the VINT device must be plugged into.
     *
     * The default is any port.
     */
    // Phidget_setHubPort(ch, 0);
    
    /*
     * Specifies that the channel should only match a VINT hub port.
     * The only valid channel id is 0.
     *
     * The default is 0 (false), meaning VINT hub ports will never match
     */
    // Phidget_setIsHubPortDevice(ch, 1);
    
    /*
     * Specifies which channel to attach to.  It is important that the channel of
     * the device is the same class as the channel that is being opened.
     *
     * The default is any channel.
     */
    // Phidget_setChannel(ch, 0);
    
    /*
     * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
     * In a normal environment this can be done automatically by enabling server discovery, which
     * will cause the client to discovery and connect to available servers.
     *
     * To force the channel to only match a network Phidget, set remote to 1.
     */
    // PhidgetNet_enableServerDiscovery(PHIDGETSERVER_DEVICEREMOTE);
    // Phidget_setIsRemote(ch, 1);
    
    return EPHIDGET_OK;
}


#pragma mark Event callbacks
static void gotAttach(PhidgetHandle phid, void *context){
    [(__bridge id)context performSelectorOnMainThread:@selector(onAttachHandler)
                                           withObject:nil
                                        waitUntilDone:NO];
}
static void gotDetach(PhidgetHandle phid, void *context){
    [(__bridge id)context performSelectorOnMainThread:@selector(onDetachHandler)
                                           withObject:nil
                                        waitUntilDone:NO];
}
static void gotError(PhidgetHandle phid, void *context, Phidget_ErrorEventCode errcode, const char *error){
    [(__bridge id)context performSelectorOnMainThread:@selector(errorHandler:)
                                           withObject:[NSArray arrayWithObjects:[NSNumber numberWithInt:errcode], [NSString stringWithUTF8String:error], nil]
                                        waitUntilDone:NO];
}
static void gotPositionChangeData(PhidgetStepperHandle phid, void *context, double position){
        [(__bridge id)context performSelectorOnMainThread:@selector(onPositionChangeHandler:)
                                               withObject:[NSNumber numberWithDouble:position]
                                            waitUntilDone:NO];
}
static void gotStoppedData(PhidgetStepperHandle phid, void *context){
        [(__bridge id)context performSelectorOnMainThread:@selector(onStoppedHandler)
                                               withObject:nil
                                            waitUntilDone:NO];
}
static void gotVelocityChangeData(PhidgetStepperHandle phid, void *context, double velocity){
        [(__bridge id)context performSelectorOnMainThread:@selector(onVelocityChangeHandler:)
                                               withObject:[NSNumber numberWithDouble:velocity]
                                            waitUntilDone:NO];
}


#pragma mark Attach, detach, data, and error events
- (void)onAttachHandler{
    PhidgetReturnCode result;
    int heightChange = 450;
    double minCurrentLimit, maxCurrentLimit, currentLimit;
    Phidget_DeviceID deviceID;
    int engaged;
    
    //Get information from channel that will allow us to configure the GUI properly
    result = Phidget_getDeviceID((PhidgetHandle)ch, &deviceID);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMaxCurrentLimit(ch, &maxCurrentLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMinCurrentLimit(ch, &minCurrentLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getCurrentLimit(ch, &currentLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getEngaged(ch, &engaged);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }

    //Adjust GUI based on information from channel
    [phidgetInfoBoxView fillPhidgetInfo:(PhidgetHandle)ch];
    [motorStatusBox setHidden:NO];
    [tabBox setHidden:NO];
    
    [self updateGUI:YES];
    
    if(deviceID != PHIDID_1062){
        [positionCurrentLimitBox setHidden:NO];
        [velocityCurrentyLimitBox setHidden:NO];
        [positionCurrentLimitSlider setMinValue:0.0];
        [positionCurrentLimitSlider setMaxValue:maxCurrentLimit];
        [positionMinCurrentLimitLabel setStringValue:@"0.00 A"];
        [positionMaxCurrentLimitLabel setStringValue:[NSString stringWithFormat:@"%.2f A",maxCurrentLimit]];
        [positionCurrentLimitSlider setDoubleValue:currentLimit];
        [positionCurrentLimitText setStringValue:[NSString stringWithFormat:@"%.2f A", currentLimit]];
        [velocityCurrentLimitSlider setMinValue:0.0];
        [velocityCurrentLimitSlider setMaxValue:maxCurrentLimit];
        [velocityMinCurrentLimitLabel setStringValue:@"0.00 A"];
        [velocityMaxCurrentLimitLabel setStringValue:[NSString stringWithFormat:@"%.2f A",maxCurrentLimit]];
        [velocityCurrentLimitSlider setDoubleValue:currentLimit];
        [velocityCurrentLimitText setStringValue:[NSString stringWithFormat:@"%.2f A", currentLimit]];
    }
    else{
        [positionCurrentLimitBox setHidden:YES];
        [velocityCurrentyLimitBox setHidden:YES];
    }
    
    if(deviceID == PHIDID_STC1000){
        [positionHoldingCurrentLimitBox setHidden:NO];
        [velocityHoldingCurrentLimitBox setHidden:NO];
    }
    else{
        [positionHoldingCurrentLimitBox setHidden:YES];
        [velocityHoldingCurrentLimitBox setHidden:YES];
    }
    
    if(engaged){
        engagedButton.title = @"Disengage";
    }
    else{
        engagedButton.title = @"Engage";
    }
    

    //Adjusting view height
    NSRect frame = [mainWindow frame];
    frame.size.width = 550;
    frame.size.height += heightChange;
    frame.origin.y -= heightChange;
    [mainWindow setFrame:frame display:NO animate:NO];
    [mainWindow center];

}

- (void)onDetachHandler{
    //Reset
    [motorPositionText setStringValue:@""];
    [motorVelocityText setStringValue:@""];
    [motorStatusBox setHidden:YES];
    [tabBox setHidden:YES];
    [positionHoldingCurrentLimitBox setHidden:YES];
    [velocityHoldingCurrentLimitBox setHidden:YES];
    [phidgetInfoBoxView fillPhidgetInfo:nil];
    
    //Adjusting window height
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
}

static int errorCounter = 0;
-(void) outputError:(const char *)errorString{
    errorCounter++;
    NSAttributedString *outputString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[NSString stringWithUTF8String:errorString]]
    	attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
    [[errorEventLog textStorage] beginEditing];
    [[errorEventLog textStorage] appendAttributedString:outputString];
    [[errorEventLog textStorage] endEditing];
    
    [errorEventLogCounter setIntValue:errorCounter];
    if(![errorEventLogWindow isVisible])
        [errorEventLogWindow setIsVisible:YES];
}

-(void)errorHandler:(NSArray *)errorEventData{
    const char* errorString = [[errorEventData objectAtIndex:1] UTF8String];
    [self outputError:errorString];
}

-(void)addError:(PhidgetReturnCode)result{
    const char *errorString;
    Phidget_getErrorDescription(result, &errorString);
    [self outputError:errorString];
}


- (IBAction)clearErrorLog:(id)sender{
    [[errorEventLog textStorage] setAttributedString:[[NSAttributedString alloc] initWithString:@""]];
    [errorEventLogCounter setIntValue:0];
    errorCounter = 0;
}

- (void)onPositionChangeHandler:(NSNumber *)position{
    [motorPositionText setStringValue:[NSString stringWithFormat:@"%.3f",[position doubleValue]]];
    [self updateStopped];
}

- (void)onVelocityChangeHandler:(NSNumber *)velocity{
    [motorVelocityText setStringValue:[NSString stringWithFormat:@"%.3f",[velocity doubleValue]]];
    [self updateStopped];
}

- (void)onStoppedHandler
{
    [motorMovingLabel setStringValue:@"Stopped"];
    [motorMovingLabel setBackgroundColor:[NSColor  colorWithRed:1.0 green:0.494 blue:0.475 alpha:1]];
}

-(void)updateStopped{
    int isMoving, engaged;
    PhidgetReturnCode result;
    result = PhidgetStepper_getEngaged(ch, &engaged);
    if(!engaged)
        return;
    result = PhidgetStepper_getIsMoving(ch, &isMoving);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    
    if(isMoving){
        [motorMovingLabel setStringValue:@"Moving"];
        [motorMovingLabel setBackgroundColor:[NSColor colorWithRed:0 green:0.561 blue:0 alpha:1]];
    }
    else{
        [motorMovingLabel setStringValue:@"Stopped"];
        [motorMovingLabel setBackgroundColor:[NSColor  colorWithRed:1.0 green:0.494 blue:0.475 alpha:1]];
    }
}

#pragma mark GUI controls

- (void)tabView:(NSTabView *)tabView
didSelectTabViewItem:(NSTabViewItem *)tabViewItem{
    PhidgetReturnCode result;
    if([tabView indexOfTabViewItem:tabViewItem] == 0){
        result = PhidgetStepper_setControlMode(ch, CONTROL_MODE_STEP);
    }
    else{
        result = PhidgetStepper_setControlMode(ch, CONTROL_MODE_RUN);
    }
    
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
}

- (IBAction)engaged:(NSButton *)sender {
    PhidgetReturnCode result;
    int engaged;
    PhidgetStepper_getEngaged(ch, &engaged);
    if(engaged){
        result = PhidgetStepper_setEngaged(ch, 0);
        if(result != EPHIDGET_OK){
            [self addError:result];
            return;
        }
        [sender setTitle:@"Engage"];
    }
    else{
        result = PhidgetStepper_setEngaged(ch, 1);
        if(result != EPHIDGET_OK){
            [self addError:result];
            return;
        }
        [sender setTitle:@"Disengage"];
    }
}


- (IBAction)setRescaleFactor:(id)sender{
    PhidgetReturnCode result;
    NSButton *field = (NSButton *)sender;
    result = PhidgetStepper_setRescaleFactor(ch, [[rescaleTextField stringValue] doubleValue]);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    
    [self updateGUI:NO];
}

//Position tab events
- (IBAction)setHoldingCurrentLimit:(id)sender {
    PhidgetReturnCode result;
    NSButton *button = (NSButton *)sender;
    double holdingCurrent;
    if(button.tag == 0)
        holdingCurrent = [positionHoldingCurrentTextField doubleValue];
    else
        holdingCurrent = [velocityHoldingCurrentLimitTextField doubleValue];
    
    result = PhidgetStepper_setHoldingCurrentLimit(ch, holdingCurrent);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    
}

- (IBAction)setRange:(id)sender {
    PhidgetReturnCode result;
    double minPosition, maxPosition;
    double absValue = fabs([positionRangeText doubleValue]);
    result = PhidgetStepper_getMinPosition(ch, &minPosition);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    result = PhidgetStepper_getMaxPosition(ch, &maxPosition);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    if(absValue > maxPosition){
        [positionRangeText setStringValue:@"Invalid"];
        return;
    }
    [positionTargetPositionSlider setMinValue:-absValue];
    [positionTargetPositionSlider setMaxValue:absValue];
    [positionMinTargetPositionLabel setStringValue:[NSString stringWithFormat:@"%.3f",-absValue]];
    [positionMaxTargetPositionLabel setStringValue:[NSString stringWithFormat:@"%.3f",absValue]];
}

- (IBAction)setTargetPosition:(id)sender{
    PhidgetReturnCode result;
    result = PhidgetStepper_setTargetPosition(ch, [positionTargetPositionSlider doubleValue]);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    [positionTargetPositionText setStringValue:[NSString stringWithFormat:@"%.3f",[positionTargetPositionSlider doubleValue]]];
    [self updateStopped];
}

- (IBAction)setCurrentLimit:(id)sender{
    PhidgetReturnCode result;
    NSSlider *slider = (NSSlider *)sender;
    result = PhidgetStepper_setCurrentLimit(ch, [slider doubleValue]);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    [positionCurrentLimitSlider setDoubleValue:[slider doubleValue]];
    [velocityCurrentLimitSlider setDoubleValue:[slider doubleValue]];
    [positionCurrentLimitText setStringValue:[NSString stringWithFormat:@"%.2f A",[slider doubleValue]]];
    [velocityCurrentLimitText setStringValue:[NSString stringWithFormat:@"%.2f A",[slider doubleValue]]];
}

- (IBAction)setAcceleration:(id)sender{
    PhidgetReturnCode result;
    NSSlider *slider = (NSSlider *)sender;
    result = PhidgetStepper_setAcceleration(ch, [slider doubleValue]);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    
    [positionAccelerationSlider setDoubleValue:[slider doubleValue]];
    [velocityAccelerationSlider setDoubleValue:[slider doubleValue]];
    [positionAccelerationText setStringValue:[NSString stringWithFormat:@"%.3f",[slider doubleValue]]];
    [velocityAccelerationText setStringValue:[NSString stringWithFormat:@"%.3f",[slider doubleValue]]];
}

- (IBAction)setVelocityLimit:(id)sender{
    PhidgetReturnCode result;
    NSSlider *slider = (NSSlider *)sender;
    result = PhidgetStepper_setVelocityLimit(ch, [slider doubleValue]);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    [positionVelocityLimitSlider setDoubleValue:[slider doubleValue]];
    [velocityVelocityLimitSlider setDoubleValue:[slider doubleValue]];
    [positionVelocityLimitText setStringValue:[NSString stringWithFormat:@"%.3f",[slider doubleValue]]];
    [velocityVelocityLimitText setStringValue:[NSString stringWithFormat:@"%.3f",[slider doubleValue]]];
}

- (IBAction)positionZero:(id)sender{
    double position;
    PhidgetReturnCode result;
    PhidgetStepper_getPosition(ch, &position);
    result = PhidgetStepper_addPositionOffset(ch, -position);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getPosition(ch, &position);
    
    [motorPositionText setStringValue:[NSString stringWithFormat:@"%.3f", position]];
}

- (IBAction)velocityStop:(id)sender{
    PhidgetReturnCode result;
    result = PhidgetStepper_setVelocityLimit(ch, 0);
    if(result != EPHIDGET_OK){
        [self addError:result];
        return;
    }
    [velocityVelocityLimitText setStringValue:@"0"];
    [velocityVelocityLimitSlider setDoubleValue:0.0];
}

#pragma mark Misc
-(void)updateGUI:(BOOL)isInitialUpdate{
    PhidgetReturnCode result;
    double minAcceleration, maxAcceleration, acceleration;
    double minVelocityLimit, maxVelocityLimit, velocityLimit, velocity;
    double minPosition, maxPosition, position;
    PhidgetStepper_ControlMode controlMode;
    double rescaleFactor;
    
    result = PhidgetStepper_getControlMode(ch, &controlMode);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMinVelocityLimit(ch, &minVelocityLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMaxVelocityLimit(ch, &maxVelocityLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getVelocityLimit(ch, &velocityLimit);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMinAcceleration(ch, &minAcceleration);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getMaxAcceleration(ch, &maxAcceleration);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getAcceleration(ch, &acceleration);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getPosition(ch, &position);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getVelocity(ch, &velocity);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }
    result = PhidgetStepper_getRescaleFactor(ch, &rescaleFactor);
    if(result != EPHIDGET_OK){
        [self addError:result];
    }

    if(controlMode == CONTROL_MODE_STEP){
        minPosition = -10000 * rescaleFactor;
        maxPosition = 10000 * rescaleFactor;
        [positionTargetPositionText setStringValue:@"0"];
        [positionTargetPositionSlider setDoubleValue:0.0];
        [positionMinTargetPositionLabel setStringValue:[NSString stringWithFormat:@"%.3f",minPosition]];
        [positionMaxTargetPositionLabel setStringValue:[NSString stringWithFormat:@"%.3f",maxPosition]];
        [positionTargetPositionSlider setMinValue:minPosition];
        [positionTargetPositionSlider setMaxValue:maxPosition];
        [positionRangeText setStringValue:[NSString stringWithFormat:@"%.3f",maxPosition]];
    }
    
    if(isInitialUpdate){
        [positionRangeText setStringValue:@"100000"];
    }
    
    [motorPositionText setStringValue:[NSString stringWithFormat:@"%.3f",position]];
    [motorVelocityText setStringValue:[NSString stringWithFormat:@"%.3f",velocity]];
    [positionAccelerationText setStringValue:[NSString stringWithFormat:@"%.3f",acceleration]];
    [positionAccelerationSlider setDoubleValue:acceleration];
    [positionMinAccelerationLabel setStringValue:[NSString stringWithFormat:@"%.3f",minAcceleration]];
    [positionMaxAccelerationLabel setStringValue:[NSString stringWithFormat:@"%.3f", maxAcceleration]];
    [positionAccelerationSlider setMinValue:minAcceleration];
    [positionAccelerationSlider setMaxValue:maxAcceleration];
    
    [positionVelocityLimitText setStringValue:[NSString stringWithFormat:@"%.3f", velocityLimit]];
    [positionVelocityLimitSlider setDoubleValue:velocityLimit];
    [positionMinVelocityLimitLabel setStringValue:[NSString stringWithFormat:@"%.3f", minVelocityLimit]];
    [positionMaxVelocityLimitLabel setStringValue:[NSString stringWithFormat:@"%.3f",maxVelocityLimit]];
    [positionVelocityLimitSlider setMinValue:minVelocityLimit];
    [positionVelocityLimitSlider setMaxValue:maxVelocityLimit];
    
    [velocityAccelerationText setStringValue:[NSString stringWithFormat:@"%.3f", acceleration]];
    [velocityAccelerationSlider setDoubleValue:acceleration];
    [velocityMinAccelerationLabel setStringValue:[NSString stringWithFormat:@"%.3f", minAcceleration]];
    [velocityMaxAccelerationLabel setStringValue:[NSString stringWithFormat:@"%.3f", maxAcceleration]];
    [velocityAccelerationSlider setMinValue:minAcceleration];
    [velocityAccelerationSlider setMaxValue:maxAcceleration];
    
    [velocityVelocityLimitText setStringValue:[NSString stringWithFormat:@"%.3f", velocityLimit]];
    [velocityVelocityLimitSlider setDoubleValue:velocityLimit];
    [velocityMaxVelocityLimitLabel setStringValue:[NSString stringWithFormat:@"%.3f", maxVelocityLimit]];
    [velocityMinVelocityLimitLabel setStringValue:[NSString stringWithFormat:@"%.3f", -1*maxVelocityLimit]];
    [velocityVelocityLimitSlider setMinValue:-1 * maxVelocityLimit];
    [velocityVelocityLimitSlider setMaxValue:maxVelocityLimit];
}

@end
